@@ -366,7 +366,7 @@ VarCorr.brmsfit <- function(x, sigma = 1, summary = TRUE, robust = FALSE,
     tmp <- list()
   }
   # include residual variances in the output as well
-  bterms <- parse_bf(x$formula)
+  bterms <- brmsterms(x$formula)
   if (is.brmsterms(bterms)) {
     if (simple_sigma(bterms) && !is.mixfamily(x$family)) {
       tmp_resid <- list(rnames = bterms$resp, sd_pars = "sigma")
@@ -439,7 +439,7 @@ nsamples.brmsfit <- function(object, subset = NULL,
 nobs.brmsfit <- function(object, resp = NULL, ...) {
   if (is_mv(object) && length(resp)) {
     resp <- validate_resp(resp, object, multiple = FALSE)
-    bterms <- parse_bf(object$formula$forms[[resp]])
+    bterms <- brmsterms(object$formula$forms[[resp]])
     out <- nrow(subset_data(model.frame(object), bterms))
   } else {
     out <- nrow(model.frame(object))
@@ -531,6 +531,7 @@ family.brmsfit <- function(object, resp = NULL, ...) {
 #' @export
 expose_functions.brmsfit <- function(x, vectorize = FALSE, 
                                      env = globalenv(), ...) {
+  require_backend("rstan", x)
   vectorize <- as_one_logical(vectorize)
   if (vectorize) {
     funs <- rstan::expose_stan_functions(x$fit, env = environment(), ...)
