@@ -51,6 +51,10 @@
 #'   or alternatively a \code{\link{cor_brms}} object (deprecated).
 #'   If \code{autocor} is specified in \code{brmsformula}, it will 
 #'   overwrite the value specified in other functions.
+#' @param unused An optional \code{formula} which contains variables
+#'   that are unused in the model but should still be stored in the
+#'   model's data frame. This can be useful, for example,
+#'   if those variables are required for post-processing the model.
 #' 
 #' @return An object of class \code{brmsformula}, which
 #'   is essentially a \code{list} containing all model
@@ -65,32 +69,38 @@
 #'   The \code{formula} argument accepts formulas of the following syntax:
 #'   
 #'   \code{response | aterms ~ pterms + (gterms | group)} 
-#'   
-#'   The \code{pterms} part contains effects that are assumed to be the 
-#'   same across observations. We call them 'population-level' effects
+#'
+#'   The \code{pterms} part contains effects that are assumed to be the same
+#'   across observations. We call them 'population-level' or 'overall' effects,
 #'   or (adopting frequentist vocabulary) 'fixed' effects. The optional
-#'   \code{gterms} part may contain effects that are assumed to vary
-#'   across grouping variables specified in \code{group}. We
-#'   call them 'group-level' effects or (adopting frequentist 
-#'   vocabulary) 'random' effects, although the latter name is misleading
-#'   in a Bayesian context. For more details type 
-#'   \code{vignette("brms_overview")} and \code{vignette("brms_multilevel")}. 
+#'   \code{gterms} part may contain effects that are assumed to vary across
+#'   grouping variables specified in \code{group}. We call them 'group-level' or
+#'   'varying' effects, or (adopting frequentist vocabulary) 'random' effects,
+#'   although the latter name is misleading in a Bayesian context. For more
+#'   details type \code{vignette("brms_overview")} and
+#'   \code{vignette("brms_multilevel")}.
 #'   
 #'   \bold{Group-level terms}
 #'   
-#'   Multiple grouping factors each with multiple group-level effects 
-#'   are possible. (Of course we can also run models without any
-#'   group-level effects.) 
-#'   Instead of \code{|} you may use \code{||} in grouping terms
-#'   to prevent correlations from being modeled. 
-#'   Alternatively, it is possible to model different group-level terms of 
-#'   the same grouping factor as correlated (even across different formulas,
-#'   e.g., in non-linear models) by using \code{|<ID>|} instead of \code{|}.
-#'   All group-level terms sharing the same ID will be modeled as correlated.
-#'   If, for instance, one specifies the terms \code{(1+x|2|g)} and 
-#'   \code{(1+z|2|g)} somewhere in the formulas passed to \code{brmsformula},
-#'   correlations between the corresponding group-level effects 
-#'   will be estimated.
+#'   Multiple grouping factors each with multiple group-level effects are
+#'   possible. (Of course we can also run models without any group-level
+#'   effects.) Instead of \code{|} you may use \code{||} in grouping terms to
+#'   prevent correlations from being modeled. Equivalently, the \code{cor}
+#'   argument of the \code{\link{gr}} function can be used for this purpose, 
+#'   for example, \code{(1 + x || g)} is equivalent to 
+#'   \code{(1 + x | gr(g, cor = FALSE))}.
+#'   
+#'   It is also possible to model different group-level terms of the same
+#'   grouping factor as correlated (even across different formulas, e.g., in
+#'   non-linear models) by using \code{|<ID>|} instead of \code{|}. All
+#'   group-level terms sharing the same ID will be modeled as correlated. If,
+#'   for instance, one specifies the terms \code{(1+x|i|g)} and \code{(1+z|i|g)}
+#'   somewhere in the formulas passed to \code{brmsformula}, correlations
+#'   between the corresponding group-level effects will be estimated. In the
+#'   above example, \code{i} is not a variable in the data but just a symbol to
+#'   indicate correlations between multiple group-level terms. Equivalently, the
+#'   \code{id} argument of the \code{\link{gr}} function can be used as well,
+#'   for example, \code{(1 + x | gr(g, id = "i"))}.
 #'   
 #'   If levels of the grouping factor belong to different sub-populations,
 #'   it may be reasonable to assume a different covariance matrix for each 
@@ -113,7 +123,7 @@
 #'   
 #'   \bold{Special predictor terms}
 #'   
-#'   Smoothing terms can modeled using the \code{\link{s}}
+#'   Flexible non-linear smooth terms can modeled using the \code{\link{s}}
 #'   and \code{\link{t2}} functions in the \code{pterms} part 
 #'   of the model formula. This allows to fit generalized additive mixed
 #'   models (GAMMs) with \pkg{brms}. The implementation is similar to that 
@@ -125,7 +135,7 @@
 #'   smooth terms, Gaussian processes can be used to model complex non-linear
 #'   relationships, for instance temporal or spatial autocorrelation. 
 #'   However, they are computationally demanding and are thus not recommended 
-#'   for very large datasets.
+#'   for very large datasets or approximations need to be used.
 #'   
 #'   The \code{pterms} and \code{gterms} parts may contain four non-standard
 #'   effect types namely monotonic, measurement error, missing value, and 
@@ -193,8 +203,8 @@
 #'   
 #'   \bold{Autocorrelation terms}
 #'   
-#'   Autocorrelation terms can be directly specified inside the formula
-#'   as well. Details can be found in \code{\link{autocor-terms}}.
+#'   Autocorrelation terms can be directly specified inside the \code{pterms}
+#'   part as well. Details can be found in \code{\link{autocor-terms}}.
 #'   
 #'   \bold{Additional response information}
 #'   
@@ -562,7 +572,10 @@
 #' bf(y ~ a1 - a2^x, a1 ~ 1, a2 ~ x + (x|g), nl = TRUE)
 #' 
 #' # correlated group-level effects across parameters
-#' bf(y ~ a1 - a2^x, a1 ~ 1 + (1|2|g), a2 ~ x + (x|2|g), nl = TRUE)
+#' bf(y ~ a1 - a2^x, a1 ~ 1 + (1 |2| g), a2 ~ x + (x |2| g), nl = TRUE)
+#' # alternative but equivalent way to specify the above model
+#' bf(y ~ a1 - a2^x, a1 ~ 1 + (1 | gr(g, id = 2)), 
+#'    a2 ~ x + (x | gr(g, id = 2)), nl = TRUE)
 #' 
 #' # define a multivariate model
 #' bf(mvbind(y1, y2) ~ x * z + (1|g))
@@ -610,6 +623,10 @@
 #'   gaussian() + cor_ar(~1|g) +
 #'   bf(y2 ~ z) + poisson()
 #'   
+#' # specify correlated residuals of a gaussian and a poisson model
+#' form1 <- bf(y1 ~ 1 + x + (1|c|obs), sigma = 1) + gaussian()
+#' form2 <- bf(y2 ~ 1 + x + (1|c|obs)) + poisson()
+#'   
 #' # model missing values in predictors
 #' bf(bmi ~ age * mi(chl)) +
 #'   bf(chl | mi() ~ age) + 
@@ -625,7 +642,7 @@
 brmsformula <- function(formula, ..., flist = NULL, family = NULL,
                         autocor = NULL, nl = NULL, loop = NULL, 
                         center = NULL, cmc = NULL, sparse = NULL,
-                        decomp = NULL) {
+                        decomp = NULL, unused = NULL) {
   if (is.brmsformula(formula)) {
     out <- formula
   } else {
@@ -709,18 +726,21 @@ brmsformula <- function(formula, ..., flist = NULL, family = NULL,
   if (!is.null(decomp)) {
     attr(out$formula, "decomp") <- match.arg(decomp, decomp_opts())
   }
+  if (!is.null(unused)) {
+    attr(out$formula, "unused") <- as.formula(unused)
+  }
   if (!is.null(autocor)) {
     attr(out$formula, "autocor") <- validate_autocor(autocor)
   } else if (!is.null(out$autocor)) {
     # for backwards compatibility with brms <= 2.11.0
     attr(out$formula, "autocor") <- validate_autocor(out$autocor)
     out$autocor <- NULL
-  } 
+  }
   if (!is.null(family)) {
-    out$family <- check_family(family)
+    out$family <- validate_family(family)
   }
   if (!is.null(lhs(formula))) {
-    out$resp <- parse_resp(formula)
+    out$resp <- terms_resp(formula)
   }
   # add default values for unspecified elements
   defs <- list(pforms = list(), pfix = list(), family = NULL, resp = NULL)
@@ -863,7 +883,7 @@ lf <- function(..., flist = NULL, dpar = NULL, resp = NULL,
 #' @rdname brmsformula-helpers
 #' @export
 acformula <- function(autocor, resp = NULL) {
-  autocor <- parse_ac(as.formula(autocor))
+  autocor <- terms_ac(as.formula(autocor))
   if (!is.formula(autocor)) {
     stop2("'autocor' must contain at least one autocorrelation term.")
   }
@@ -958,7 +978,7 @@ mvbf <- function(..., flist = NULL, rescor = NULL) {
 # which uses mvbind on the left-hand side to specify MV models
 split_bf <- function(x) {
   stopifnot(is.brmsformula(x))
-  resp <- parse_resp(x$formula, check_names = FALSE)
+  resp <- terms_resp(x$formula, check_names = FALSE)
   str_adform <- formula2str(x$formula)
   str_adform <- get_matches("\\|[^~]*(?=~)", str_adform, perl = TRUE)
   if (length(resp) > 1L) {
@@ -1007,8 +1027,10 @@ allow_rescor <- function(x) {
   parts <- if (is.mvbrmsformula(x)) x$forms else x$terms 
   families <- lapply(parts, "[[", "family")
   has_rescor <- ulapply(families, has_rescor)
+  is_mixture <- ulapply(families, is.mixfamily)
   family_names <- ulapply(families, "[[", "family")
-  all(has_rescor) && length(unique(family_names)) == 1L
+  all(has_rescor) && !any(is_mixture) &&
+    length(unique(family_names)) == 1L
 }
 
 #' @rdname brmsformula-helpers
@@ -1215,32 +1237,89 @@ validate_formula.default <- function(formula, ...) {
 # @param formula object of class 'formula' of 'brmsformula'
 # @param data optional data.frame to validate data related arguments
 # @param family optional 'family' object
-# @param autocor optional 'cor_brms' object
+# @param autocor (deprecated) optional 'cor_brms' object
 # @param threshold (deprecated) threshold type for ordinal models
+# @param cov_ranef (deprecated) named list of group covariance matrices
 # @return a brmsformula object compatible with the current version of brms
 #' @export
 validate_formula.brmsformula <- function(
   formula, family = gaussian(), autocor = NULL, 
-  data = NULL, threshold = NULL, sparse = NULL, ...
+  data = NULL, threshold = NULL, sparse = NULL,
+  cov_ranef = NULL, ...
 ) {
   out <- bf(formula)
   if (is.null(out$family) && !is.null(family)) {
-    out$family <- check_family(family)
-  }
-  if (is.null(attr(out$formula, "autocor")) && !is.null(autocor)) {
-    # 'autocor' interface has been changed in brms 2.11.1
-    warning2(
-      "Argument 'autocor' should be specified within the ", 
-      "'formula' argument. See ?brmsformula for help."
-    )
-    # store 'autocor' as an attribute to carry it around easier
-    attr(out$formula, "autocor") <- validate_autocor(autocor)
+    out$family <- validate_family(family)
   }
   # allow the '.' symbol in the formulas
   out$formula <- expand_dot_formula(out$formula, data)
   for (i in seq_along(out$pforms)) {
     out$pforms[[i]] <- expand_dot_formula(out$pforms[[i]], data)
   }
+  # allow 'me' terms to be correlated
+  out$mecor <- default_mecor(out$mecor)
+  if (has_cat(out) && !is.null(data)) {
+    # for easy access of response categories
+    # allow to update 'cats' with new data
+    out$family$cats <- extract_cat_names(out, data)
+  }
+  if (is_ordinal(out$family)) {
+    # thresholds and category names are data dependent
+    try_terms <- try(stats::terms(out$formula), silent = TRUE)
+    intercept <- attr(try_terms, "intercept", TRUE)
+    if (!is(try_terms, "try-error") && isTRUE(intercept == 0)) {
+      stop2("Cannot remove the intercept in an ordinal model.")
+    }
+    if (!is.null(data)) {
+      # for easy access of thresholds and response categories (#838)
+      # allow to update 'cats' and 'thres' with new data
+      out$family$thres <- extract_thres_names(out, data) 
+      out$family$cats <- extract_cat_names(out, data)
+    }
+    if (is.mixfamily(out$family)) {
+      # every mixture family needs to know about response categories
+      for (i in seq_along(out$family$mix)) {
+        out$family$mix[[i]]$thres <- out$family$thres
+      }
+    }
+  }
+  conv_cats_dpars <- conv_cats_dpars(out$family)
+  if (conv_cats_dpars && !is.null(data)) {
+    # allow to update 'dpars' with new data
+    # define distributional parameters based on response categories
+    if (length(out$family$cats) < 2L) {
+      stop2("At least 2 response categories are required.")
+    }
+    if (is.null(out$family$refcat)) {
+      # the first level serves as the reference category
+      out$family$refcat <- out$family$cats[1]
+    }
+    if (isNA(out$family$refcat)) {
+      # implies predicting all categories
+      predcats <- out$family$cats
+    } else {
+      if (!out$family$refcat %in% out$family$cats) {
+        stop2("The reference response category must be one of ",
+              collapse_comma(out$family$cats), ".")
+      }
+      predcats <- setdiff(out$family$cats, out$family$refcat)
+    }
+    mu_dpars <- make_stan_names(paste0("mu", predcats))
+    if (any(duplicated(mu_dpars))) {
+      stop2("Invalid response category names. Please avoid ",
+            "using any special characters in the names.")
+    }
+    old_mu_dpars <- str_subset(out$family$dpars, "^mu")
+    out$family$dpars <- setdiff(out$family$dpars, old_mu_dpars)
+    out$family$dpars <- union(mu_dpars, out$family$dpars)
+  }
+  
+  # incorporate deprecated arguments
+  require_threshold <- is_ordinal(out$family) && is.null(out$family$threshold)
+  if (require_threshold && !is.null(threshold)) {
+    # slot 'threshold' is deprecated as of brms 1.7.0
+    out$family <- validate_family(out$family, threshold = threshold)
+  }
   if (!is.null(sparse)) {
     # a global 'sparse' argument is deprecated as of brms 2.8.3
     warning2(
@@ -1257,62 +1336,18 @@ validate_formula.brmsformula <- function(
       }
     }
   }
-  out$mecor <- default_mecor(out$mecor)
-  if (has_cat(out) && is.null(get_cats(out)) && !is.null(data)) {
-    # for easy access of response categories
-    out$family$cats <- extract_cat_names(out, data)
+  if (is.null(attr(out$formula, "autocor")) && !is.null(autocor)) {
+    # 'autocor' interface has been changed in brms 2.11.1
+    warning2(
+      "Argument 'autocor' should be specified within the ", 
+      "'formula' argument. See ?brmsformula for help."
+    )
+    # store 'autocor' as an attribute to carry it around more easily
+    attr(out$formula, "autocor") <- validate_autocor(autocor)
   }
-  if (is_ordinal(out$family)) {
-    if (is.null(out$family$threshold) && !is.null(threshold)) {
-      # slot 'threshold' is deprecated as of brms 1.7.0
-      out$family <- check_family(out$family, threshold = threshold)
-    }
-    try_terms <- try(stats::terms(out$formula), silent = TRUE)
-    intercept <- attr(try_terms, "intercept", TRUE)
-    if (!is(try_terms, "try-error") && isTRUE(intercept == 0)) {
-      stop2("Cannot remove the intercept in an ordinal model.")
-    }
-    if (is.null(get_thres(out)) && !is.null(data)) {
-      # for easy access of thresholds
-      out$family$thres <- extract_thres_names(out, data)  
-    }
-    if (is.null(get_cats(out)) && !is.null(data)) {
-      # for easy access of response categories (#838)
-      out$family$cats <- extract_cat_names(out, data)
-    }
-    if (is.mixfamily(out$family)) {
-      # every mixture family needs to know about response categories
-      for (i in seq_along(out$family$mix)) {
-        out$family$mix[[i]]$thres <- out$family$thres
-      }
-    }
-  }
-  mu_dpars <- str_subset(out$family$dpars, "^mu")
-  conv_cats_dpars <- conv_cats_dpars(out$family)
-  if (conv_cats_dpars && !length(mu_dpars) && !is.null(data)) {
-    # define distributional parameters based on response categories
-    if (length(out$family$cats) < 2L) {
-      stop2("At least 2 response categories are required.")
-    }
-    if (is.null(out$family$refcat)) {
-      # the first level serves as the reference category
-      out$family$refcat <- out$family$cats[1]
-    } 
-    if (isNA(out$family$refcat)) {
-      predcats <- out$family$cats  # predict all categories
-    } else {
-      if (!out$family$refcat %in% out$family$cats) {
-        stop2("The reference response category must be one of ",
-              collapse_comma(out$family$cats), ".")
-      }
-      predcats <- setdiff(out$family$cats, out$family$refcat)
-    }
-    mu_dpars <- make_stan_names(paste0("mu", predcats))
-    if (any(duplicated(mu_dpars))) {
-      stop2("Invalid response category names. Please avoid ",
-            "using any special characters in the names.")
-    }
-    out$family$dpars <- c(mu_dpars, out$family$dpars)
+  if (!is.null(cov_ranef)) {
+    # 'cov_ranef' is deprecated as of brms 2.12.5
+    out$cov_ranef <- validate_cov_ranef(cov_ranef)
   }
   bf(out)
 }
@@ -1321,7 +1356,7 @@ validate_formula.brmsformula <- function(
 # allow passing lists of families or autocors
 #' @export
 validate_formula.mvbrmsformula <- function(
-  formula, family = NULL, autocor = NULL, ...
+  formula, family = NULL, autocor = NULL, cov_ranef = NULL, ...
 ) {
   nresp <- length(formula$forms)
   if (!is(family, "list")) {
@@ -1349,10 +1384,17 @@ validate_formula.mvbrmsformula <- function(
   if (is.null(formula$rescor)) {
     # with 'mi' terms we usually don't want rescor to be estimated
     miforms <- ulapply(formula$forms, function(f)
-      parse_ad(f$formula, f$family, FALSE)[["mi"]]
+      terms_ad(f$formula, f$family, FALSE)[["mi"]]
     )
     formula$rescor <- allow_rescor && !length(miforms)
     message("Setting 'rescor' to ", formula$rescor, " by default for this model")
+    if (formula$rescor) {
+      warning2(
+        "In the future, 'rescor' will be set to FALSE by default for ", 
+        "all models. It is thus recommended to explicitely set ",
+        "'rescor' via 'set_recor' instead of using the default."
+      ) 
+    }
   }
   formula$rescor <- as_one_logical(formula$rescor)
   if (formula$rescor) {
@@ -1366,6 +1408,11 @@ validate_formula.mvbrmsformula <- function(
   for (i in seq_along(formula$forms)) {
     formula$forms[[i]]$mecor <- formula$mecor
   }
+  # incorporate deprecated arguments
+  if (!is.null(cov_ranef)) {
+    # 'cov_ranef' is deprecated as of brms 2.12.5
+    formula$cov_ranef <- validate_cov_ranef(cov_ranef)
+  }
   formula
 }
 
