@@ -6,6 +6,7 @@
 #' @param resp Optional names of response variables for which to extract values.
 #' @param warn For internal use only.
 #' @param ... Further arguments passed to \code{\link{standata}}.
+#' @inheritParams posterior_predict.brmsfit
 #' 
 #' @return Returns a vector of response values for univariate models and a
 #'   matrix of response values with one column per response variable for
@@ -13,9 +14,10 @@
 #' 
 #' @keywords internal
 #' @export
-get_y <- function(x, resp = NULL, warn = FALSE, ...) {
+get_y <- function(x, resp = NULL, sort = FALSE, warn = FALSE,  ...) {
   stopifnot(is.brmsfit(x))
   resp <- validate_resp(resp, x)
+  sort <- as_one_logical(sort)
   warn <- as_one_logical(warn)
   args <- list(x, resp = resp, ...)
   args$re_formula <- NA
@@ -35,7 +37,12 @@ get_y <- function(x, resp = NULL, warn = FALSE, ...) {
   } else {
     out <- sdata[[Ynames]]
   }
-  structure(out, old_order = attr(sdata, "old_order"))
+  old_order <- attr(sdata, "old_order")
+  if (!is.null(old_order) && !sort) {
+    stopifnot(length(old_order) == NROW(out))
+    out <- p(out, old_order)
+  }
+  out
 }
 
 #' Prepare Response Data
@@ -55,11 +62,11 @@ data_response <- function(x, ...) {
 }
 
 #' @export
-data_response.mvbrmsterms <- function(x, old_sdata = NULL, ...) {
+data_response.mvbrmsterms <- function(x, basis = NULL, ...) {
   out <- list()
   for (i in seq_along(x$terms)) {
-    od <- old_sdata[[x$responses[i]]]
-    c(out) <- data_response(x$terms[[i]], old_sdata = od, ...)
+    bs <- basis$resps[[x$responses[i]]]
+    c(out) <- data_response(x$terms[[i]], basis = bs, ...)
   }
   if (x$rescor) {
     out$nresp <- length(x$responses)
@@ -70,14 +77,14 @@ data_response.mvbrmsterms <- function(x, old_sdata = NULL, ...) {
 
 #' @export
 data_response.brmsterms <- function(x, data, check_response = TRUE,
-                                    internal = FALSE, old_sdata = NULL, ...) {
-  # prepare data for the response variable
+                                    internal = FALSE, basis = NULL, ...) {
   data <- subset_data(data, x)
   N <- nrow(data)
+  # TODO: rename 'Y' to 'y'
   Y <- model.response(model.frame(x$respform, data, na.action = na.pass))
   out <- list(N = N, Y = unname(Y))
   if (is_binary(x$family) || is_categorical(x$family)) {
-    out$Y <- as_factor(out$Y, levels = old_sdata$resp_levels)
+    out$Y <- as_factor(out$Y, levels = basis$resp_levels)
     out$Y <- as.numeric(out$Y)
     if (is_binary(x$family)) {
       out$Y <- out$Y - 1
@@ -159,8 +166,8 @@ data_response.brmsterms <- function(x, data, check_response = TRUE,
           "Using 'binomial' families without specifying 'trials' ", 
           "on the left-hand side of the model formula is deprecated."
         )
-      } else if (!is.null(old_sdata$trials)) {
-        trials <- max(old_sdata$trials)
+      } else if (!is.null(basis$trials)) {
+        trials <- max(basis$trials)
       } else {
         stop2("Could not compute the number of trials.")
       }
@@ -231,7 +238,7 @@ data_response.brmsterms <- function(x, data, check_response = TRUE,
       Kthres_cumsum <- cumsum(nthres)
       Kthres_start <- c(1, Kthres_cumsum[-length(nthres)] + 1)
       Kthres_end <- Kthres_cumsum
-      Jthres <- cbind(Kthres_start, Kthres_end)[Jgrthres, ]
+      Jthres <- cbind(Kthres_start, Kthres_end)[Jgrthres, , drop = FALSE]
       out$Jthres <- Jthres
     } else {
       nthres <- max(thres$thres)
@@ -369,6 +376,11 @@ data_response.brmsterms <- function(x, data, check_response = TRUE,
         out$noise[which_mi] <- Inf
       }
     }
+    # bounds are required for predicting new missing values
+    # not required in Stan right now as bounds are hard-coded there
+    tbounds <- trunc_bounds(x, data, incl_family = TRUE)
+    out$lbmi <- tbounds$lb
+    out$ubmi <- tbounds$ub
     if (!internal) {
       # Stan does not allow NAs in data
       # use Inf to that min(Y) is not affected
@@ -412,7 +424,7 @@ data_response.brmsterms <- function(x, data, check_response = TRUE,
 }
 
 # data specific for mixture models
-data_mixture <- function(bterms, prior = brmsprior()) {
+data_mixture <- function(bterms, data2, prior) {
   stopifnot(is.brmsterms(bterms))
   out <- list()
   if (is.mixfamily(bterms$family)) {
@@ -422,16 +434,8 @@ data_mixture <- function(bterms, prior = brmsprior()) {
       # estimate mixture probabilities directly
       take <- find_rows(prior, class = "theta", resp = bterms$resp)
       theta_prior <- prior$prior[take]
-      if (isTRUE(nzchar(theta_prior))) {
-        theta_prior <- eval_dirichlet(theta_prior)
-        if (length(theta_prior) != length(families)) {
-          stop2("Invalid dirichlet prior for the ", 
-                "mixture probabilities 'theta'.")
-        }
-        out$con_theta <- theta_prior
-      } else {
-        out$con_theta <- rep(1, length(families)) 
-      }
+      con_theta <- eval_dirichlet(theta_prior, length(families), data2)
+      out$con_theta <- as.array(con_theta)
       p <- usc(combine_prefix(bterms))
       names(out) <- paste0(names(out), p)
     }
@@ -440,16 +444,20 @@ data_mixture <- function(bterms, prior = brmsprior()) {
 }
 
 # data for the baseline functions of Cox models
-data_bhaz <- function(bterms, data, basis = NULL) {
+data_bhaz <- function(bterms, data, data2, prior, basis = NULL) {
   out <- list()
   if (!is_cox(bterms$family)) {
     return(out) 
   }
   y <- model.response(model.frame(bterms$respform, data, na.action = na.pass))
   args <- bterms$family$bhaz 
-  out$Zbhaz <- bhaz_basis_matrix(y, args, basis = basis)
-  out$Zcbhaz <- bhaz_basis_matrix(y, args, integrate = TRUE, basis = basis)
+  bs <- basis$basis_matrix
+  out$Zbhaz <- bhaz_basis_matrix(y, args, basis = bs) 
+  out$Zcbhaz <- bhaz_basis_matrix(y, args, integrate = TRUE, basis = bs)
   out$Kbhaz <- NCOL(out$Zbhaz)
+  sbhaz_prior <- subset2(prior, class = "sbhaz", resp = bterms$resp)
+  con_sbhaz <- eval_dirichlet(sbhaz_prior$prior, out$Kbhaz, data2)
+  out$con_sbhaz <- as.array(con_sbhaz)
   out
 }
 
@@ -525,7 +533,7 @@ extract_thres_names <- function(x, data) {
   stopifnot(is.brmsformula(x) || is.brmsterms(x), has_thres(x))
   
   if (is.null(x$adforms)) {
-    x$adforms <- parse_ad(x$formula, x$family)
+    x$adforms <- terms_ad(x$formula, x$family)
   }
   nthres <- get_ad_values(x, "thres", "thres", data)
   if (any(!is_wholenumber(nthres) | nthres < 1L)) {
