@@ -88,8 +88,10 @@ test_that(paste("make_standata accepts correct response variables",
                as.array(rep(1:10,5)))
   expect_equal(make_standata(y ~ 1, data = data.frame(y = seq(1,10,0.1)), 
                              family = "exponential")$Y, as.array(seq(1,10,0.1)))
+  
   dat <- data.frame(y1 = 1:10, y2 = 11:20, x = rep(0,10))
-  sdata <- make_standata(mvbind(y1, y2) ~ x, data = dat)
+  form <- bf(mvbind(y1, y2) ~ x) + set_rescor(TRUE)
+  sdata <- make_standata(form, data = dat)
   expect_equal(sdata$Y_y1, as.array(1:10))
   expect_equal(sdata$Y_y2, as.array(11:20))
 })
@@ -189,15 +191,18 @@ test_that("make_standata handles multivariate models", {
     tim = 10:1, w = 1:10
   )
   
-  sdata <- make_standata(mvbind(y1, y2) | weights(w) ~ x, data = dat)
+  form <- bf(mvbind(y1, y2) | weights(w) ~ x) + set_rescor(TRUE)
+  sdata <- make_standata(form, data = dat)
   expect_equal(sdata$Y_y1, as.array(dat$y1))
   expect_equal(sdata$Y_y2, as.array(dat$y2))
   expect_equal(sdata$weights_y1, as.array(1:10))
   
-  expect_error(make_standata(mvbind(y1, y2, y2) ~ x, data = dat),
+  expect_error(make_standata(bf(mvbind(y1, y2, y2) ~ x) + set_resor(FALSE), 
+                             data = dat),
                "Cannot use the same response variable twice")
   
-  sdata <- make_standata(mvbind(y1 / y2, y2, y1 * 3) ~ x, data = dat)
+  form <- bf(mvbind(y1 / y2, y2, y1 * 3) ~ x) + set_rescor(FALSE)
+  sdata <- make_standata(form, data = dat)
   expect_equal(sdata$Y_y1y2, as.array(dat$y1 / dat$y2))
   
   sdata <- suppressWarnings(
@@ -225,18 +230,40 @@ test_that("make_standata handles multivariate models", {
     "Zs_y2_1_1", "Y_y2", "con_theta_x", "X_mu2_x"
   )
   expect_true(all(sdata_names %in% names(sdata)))
-  expect_equal(sdata$con_theta_x, c(2, 1))
+  expect_equal(sdata$con_theta_x, as.array(c(2, 1)))
+})
+
+test_that("make_standata removes NAs correctly", {
+  dat <- data.frame(y = c(rnorm(9), NA))
+  sdata <- suppressWarnings(make_standata(y ~ 1, dat))
+  expect_equal(as.numeric(sdata$Y), dat$y[1:9])
+})
+
+test_that("make_standata handles the 'subset' addition argument correctly", {
+  dat1 <- data.frame(
+    y1 = rnorm(15), y2 = NA,
+    x1 = rnorm(15), x2 = NA, x3 = rnorm(15),
+    sub1 = 1, sub2 = 0
+  )
+  dat2 <- data.frame(
+    y1 = NA, y2 = rnorm(10),
+    x1 = NA, x2 = rnorm(10), x3 = NA,
+    sub1 = 0, sub2 = 1
+  )
+  dat <- rbind(dat1, dat2)
+  
+  bform <- 
+    bf(y1 | subset(sub1) ~ x1*x3 + sin(x1), family = gaussian()) +
+    bf(y2 | subset(sub2) ~ x2, family = gaussian()) +
+    set_rescor(FALSE)
   
-  # test addition argument 'subset'
-  bform <- bf(y1 | subset(censi) ~ x + y2 + (x|2|g)) + 
-    (bf(y2 ~ s(y2) + (1|2|g)) + skew_normal())
   sdata <- make_standata(bform, dat)
-  nsub <- sum(dat$censi)
-  expect_equal(sdata$N_y1, nsub)
-  expect_equal(sdata$N_y2, nrow(dat))
-  expect_equal(length(sdata$Y_y1), nsub)
-  expect_equal(nrow(sdata$X_y1), nsub)
-  expect_equal(length(sdata$Z_1_y1_2), nsub)
+  nsub1 <- sum(dat$sub1)
+  nsub2 <- sum(dat$sub2)
+  expect_equal(sdata$N_y1, nsub1)
+  expect_equal(sdata$N_y2, nsub2)
+  expect_equal(length(sdata$Y_y1), nsub1)
+  expect_equal(nrow(sdata$X_y2), nsub2)
 })
 
 test_that("make_standata returns correct data for ARMA terms", {
@@ -261,38 +288,49 @@ test_that("make_standata allows to retrieve the initial data order", {
                           id = sample(1:10, 100, TRUE), 
                           time = sample(1:100, 100))
   # univariate model
-  sdata1 <- make_standata(y1 ~ ar(time, id), data = dat,
-                          control = list(save_order = TRUE))
+  sdata1 <- make_standata(y1 ~ ar(time, id), data = dat, internal = TRUE)
   expect_equal(dat$y1, as.numeric(sdata1$Y[attr(sdata1, "old_order")]))
   
   # multivariate model
-  sdata2 <- make_standata(mvbind(y1, y2) ~ ma(time, id), data = dat, 
-                          control = list(save_order = TRUE))
+  form <- bf(mvbind(y1, y2) ~ ma(time, id)) + set_rescor(FALSE)
+  sdata2 <- make_standata(form, data = dat, internal = TRUE)
   expect_equal(sdata2$Y_y1[attr(sdata2, "old_order")], as.array(dat$y1))
   expect_equal(sdata2$Y_y2[attr(sdata2, "old_order")], as.array(dat$y2))
 })
 
 test_that("make_standata handles covariance matrices correctly", {
   A <- structure(diag(1, 4), dimnames = list(1:4, NULL))
-  expect_equivalent(make_standata(count ~ Trt + (1|visit), data = epilepsy,
-                                  cov_ranef = list(visit = A))$Lcov_1, A)
-  B <- diag(1, 4)
-  expect_error(make_standata(count ~ Trt + (1|visit), data = epilepsy,
-                             cov_ranef = list(visit = B)),
-               "Row names are required")
-  B <- structure(diag(1, 4), dimnames = list(2:5, NULL))
-  expect_error(make_standata(count ~ Trt + (1|visit), data = epilepsy,
-                             cov_ranef = list(visit = B)),
-               "Row names .* do not match")
+  sdata <- make_standata(count ~ Trt + (1|gr(visit, cov = A)),
+                         data = epilepsy, data2 = list(A = A))
+  expect_equivalent(sdata$Lcov_1, t(chol(A)))
+  
   B <- structure(diag(1:5), dimnames = list(c(1,5,2,4,3), NULL))
-  expect_equivalent(make_standata(count ~ Trt + (1|visit), data = epilepsy,
-                             cov_ranef = list(visit = B))$Lcov_1,
-                    t(chol(B[c(1,3,5,4), c(1,3,5,4)])))
+  sdata <- make_standata(count ~ Trt + (1|gr(visit, cov = B)),
+                         data = epilepsy, data2 = list(B = B))
+  expect_equivalent(sdata$Lcov_1, t(chol(B[c(1,3,5,4), c(1,3,5,4)])))
+  
+  B <- diag(1, 4)
+  expect_error(make_standata(count ~ Trt + (1|gr(visit, cov = B)), 
+                             data = epilepsy, data2 = list(B = B)),
+               "Row or column names are required")
+  
+  B <- structure(diag(1, 4), dimnames = list(2:5, NULL))
+  expect_error(make_standata(count ~ Trt + (1|gr(visit, cov = B)), 
+                             data = epilepsy, data2 = list(B = B)),
+               "Levels of .* do not match")
+
   B <- A
   B[1,2] <- 0.5
-  expect_error(make_standata(count ~ Trt + (1|visit), data = epilepsy,
-                             cov_ranef = list(visit = B)),
-               "not symmetric")
+  expect_error(make_standata(count ~ Trt + (1|gr(visit, cov = B)), 
+                             data = epilepsy,  data2 = list(B = B)),
+               "must be symmetric")
+  
+  expect_warning(
+    sdata <- make_standata(count ~ Trt + (1|visit), data = epilepsy,
+                           cov_ranef = list(visit = A)),
+    "Argument 'cov_ranef' is deprecated"
+  )
+  expect_equivalent(sdata$Lcov_1, t(chol(A)))
 })
 
 test_that("make_standata correctly prepares data for non-linear models", {
@@ -351,6 +389,15 @@ test_that("make_standata correctly prepares data for monotonic effects", {
   expect_equal(sdata$con_simo_1, as.array(c(1, 0.5, 2)))
   expect_equal(sdata$con_simo_3, as.array(c(1, 0.5, 2)))
   
+  # test issue #924 (conditional monotonicity)
+  prior <- c(prior(dirichlet(c(1, 0.5, 2)), simo, coef = "v"),
+             prior(dirichlet(c(1,3)), simo, coef = "w"))
+  sdata <- make_standata(y ~ y*mo(x1, id = "v")*mo(x2, id = "w"), 
+                         data, prior = prior)
+  expect_equal(sdata$con_simo_1, as.array(c(1, 0.5, 2)))
+  expect_equal(sdata$con_simo_2, as.array(c(1, 3)))
+  expect_true(!"sdata$con_simo_3" %in% names(sdata))
+  
   expect_error(
     make_standata(y ~ mo(z), data = data),
     "Monotonic predictors must be integers or ordered factors"
@@ -359,8 +406,7 @@ test_that("make_standata correctly prepares data for monotonic effects", {
   prior <- c(set_prior("dirichlet(c(1,0.5,2))", class = "simo", coef = "mox21"))
   expect_error(
     make_standata(y ~ mo(x2), data = data, prior = prior),
-    "Invalid Dirichlet prior for the simplex of coefficient 'mox21'", 
-    fixed = TRUE
+    "Invalid Dirichlet prior"
   )
 })
 
@@ -451,7 +497,7 @@ test_that("make_standata handles category specific effects", {
                          data = inhaler, family = cratio())
   expect_equivalent(sdata$Z_1_4, as.array(inhaler$treat))
   expect_error(make_standata(rating ~ 1 + cse(treat), data = inhaler,
-                             family = "cumulative"), "require families")
+                             family = "cumulative"), "not supported")
   expect_error(make_standata(rating ~ 1 + (treat + cse(1)|subject), 
                              data = inhaler, family = "cratio"), 
                "category specific effects in separate group-level terms")
@@ -576,18 +622,28 @@ test_that("make_standata handles multi-membership models", {
 
 test_that("by variables in grouping terms are handled correctly", {
   gvar <- c("1A", "1B", "2A", "2B", "3A", "3B", "10", "100", "2", "3")
+  gvar <- rep(gvar, each = 10)
+  g_order <- order(gvar)
+  byvar <- c(0, 4.5, 3, 2, "x 1")
+  byvar <- factor(rep(byvar, each = 20))
   dat <- data.frame(
     y = rnorm(100), x = rnorm(100),
-    g = rep(gvar, each = 10),
-    z = factor(rep(c(0, 4.5, 3, 2, "x 1"), each = 20)),
-    z2 = factor(1:2)
+    g = gvar,
+    g2 = gvar[g_order],
+    z = byvar,
+    z2 = byvar[g_order],
+    z3 = factor(1:2)
   )
   sdata <- make_standata(y ~ x + (x | gr(g, by = z)), dat)
   expect_equal(sdata$Nby_1, 5)
   expect_equal(sdata$Jby_1, as.array(c(2, 2, 1, 1, 5, 4, 4, 5, 3, 3)))
   
-  expect_error(make_standata(y ~ x + (1|gr(g, by = z2)), dat),
-               "Some levels of 'g' correspond to multiple levels of 'z2'")
+  sdata <- make_standata(y ~ x + (x | mm(g, g2, by = cbind(z, z2))), dat)
+  expect_equal(sdata$Nby_1, 5)
+  expect_equal(sdata$Jby_1, as.array(c(2, 2, 1, 1, 5, 4, 4, 5, 3, 3)))
+  
+  expect_error(make_standata(y ~ x + (1|gr(g, by = z3)), dat),
+               "Some levels of 'g' correspond to multiple levels of 'z3'")
 })
 
 test_that("make_standata handles calls to the 'poly' function", {
@@ -635,13 +691,13 @@ test_that("make_standata includes data for mixture models", {
   data <- data.frame(y = rnorm(10), x = rnorm(10), c = 1)
   form <- bf(y ~ x, mu1 ~ 1, family = mixture(gaussian, gaussian))
   sdata <- make_standata(form, data)
-  expect_equal(sdata$con_theta, c(1, 1))
+  expect_equal(sdata$con_theta, as.array(c(1, 1)))
   expect_equal(dim(sdata$X_mu1), c(10, 1))
   expect_equal(dim(sdata$X_mu2), c(10, 2))
   
   form <- bf(y ~ x, family = mixture(gaussian, gaussian))
   sdata <- make_standata(form, data, prior = prior(dirichlet(10, 2), theta))
-  expect_equal(sdata$con_theta, c(10, 2))
+  expect_equal(sdata$con_theta, as.array(c(10, 2)))
   
   form <- bf(y ~ x, theta1 = 1, theta2 = 3, family = mixture(gaussian, gaussian))
   sdata <- make_standata(form, data)
@@ -650,19 +706,19 @@ test_that("make_standata includes data for mixture models", {
 })
 
 test_that("make_standata includes data for Gaussian processes", {
-  dat <- data.frame(y = rnorm(10), x1 = sample(1:10, 10),
+  dat <- data.frame(y = rnorm(10), x1 = rnorm(10),
                     z = factor(c(2, 2, 2, 3, 4, rep(5, 5))))
   sdata <- make_standata(y ~ gp(x1), dat)
   expect_equal(max(sdata$Xgp_1) - min(sdata$Xgp_1), 1) 
   sdata <- make_standata(y ~ gp(x1, scale = FALSE), dat)
-  expect_equal(max(sdata$Xgp_1) - min(sdata$Xgp_1), 9)
+  expect_equal(max(sdata$Xgp_1) - min(sdata$Xgp_1), max(dat$x1) - min(dat$x1))
   
-  sdata <- make_standata(y ~ gp(x1, by = z, gr = TRUE), dat)
+  sdata <- SW(make_standata(y ~ gp(x1, by = z, gr = TRUE, scale = FALSE), dat))
   expect_equal(sdata$Igp_1_2, as.array(4))
   expect_equal(sdata$Jgp_1_4, as.array(1:5))
   expect_equal(sdata$Igp_1_4, as.array(6:10))
   
-  sdata <- make_standata(y ~ gp(x1, by = y, gr = TRUE), dat)
+  sdata <- SW(make_standata(y ~ gp(x1, by = y, gr = TRUE), dat))
   expect_equal(sdata$Cgp_1, as.array(dat$y))
 })
 
@@ -675,7 +731,7 @@ test_that("make_standata includes data for approximate Gaussian processes", {
   expect_equal(dim(sdata$Xgp_1), c(10, 5))
   expect_equal(dim(sdata$slambda_1), c(5, 1))
   
-  sdata <- make_standata(y ~ gp(x1, by = z, k = 5, c = 5/4), dat)
+  sdata <- SW(make_standata(y ~ gp(x1, by = z, k = 5, c = 5/4, scale = FALSE), dat))
   expect_equal(sdata$Igp_1_2, as.array(4))
   expect_equal(sdata$Cgp_1_2, as.array(1))
   expect_equal(sdata$Igp_1_4, as.array(6:10))
@@ -864,9 +920,11 @@ test_that("data for multinomial and dirichlet models is correct", {
 test_that("make_standata handles cox models correctly", {
   data <- data.frame(y = rexp(100), x = rnorm(100))
   bform <- bf(y ~ x)
-  sdata <- make_standata(bform, data, brmsfamily("cox"))
-  expect_equal(dim(sdata$Zbhaz), c(100, 4))
-  expect_equal(dim(sdata$Zcbhaz), c(100, 4))
+  bprior <- prior(dirichlet(3), sbhaz)
+  sdata <- make_standata(bform, data, brmsfamily("cox"), prior = bprior)
+  expect_equal(dim(sdata$Zbhaz), c(100, 5))
+  expect_equal(dim(sdata$Zcbhaz), c(100, 5))
+  expect_equal(sdata$con_sbhaz, as.array(rep(3, 5)))
   
   sdata <- make_standata(bform, data, brmsfamily("cox", bhaz = list(df = 6)))
   expect_equal(dim(sdata$Zbhaz), c(100, 6))
@@ -914,3 +972,9 @@ test_that("make_standata handles grouped ordinal thresholds correctly", {
   expect_equal(sdata$nthres, as.array(c(6, 6)))
   expect_equal(sdata$ngrthres, 2)
 })
+
+test_that("information for threading is handled correctly", {
+  dat <- data.frame(y = 1:10)
+  sdata <- make_standata(y ~ 1, dat, threads = threading(2, grainsize = 3))
+  expect_equal(sdata$grainsize, 3)
+})
